/*  Copyright (C) MOXA Inc. All rights reserved.

    This software is distributed under the terms of the
    MOXA License.  See the file COPYING-MOXA for details.
*/
/*
    driver.h

    This is a header file for a generic device driver.

    2009-10-15	CF Lin
		new release
*/
#ifndef _DRIVER_H_   
#define _DRIVER_H_

#include "os_dependent.h"
#include "driver_dependent.h"
#include "buffer.h"
#include "list.h"
/* wade add */
#include "debug.h"

#define MAX_USERS_PER_PORT				1	/* define the number of concurrent users using a hardware port */

#ifdef __cplusplus
extern "C" {
#endif

/* bit-level interrupt types */
#define INTERRUPT_TYPE_RX				1
#define INTERRUPT_TYPE_TX				(1<<1)
#define INTERRUPT_TYPE_ERR				(1<<2)

/* definitions of control codes */
#define IOCTL_SET_READ_TIMEOUT			1
#define IOCTL_SET_WRITE_TIMEOUT			2
#define IOCTL_GET_IQUEUED_BYTES			3
#define IOCTL_GET_OQUEUED_BYTES			4
#define IOCTL_PURGE_INPUT_BUFFER		5
#define IOCTL_PURGE_OUTPUT_BUFFER		6

/* interface */
typedef struct _HWPORT HWPORT;

typedef struct _IRQMEM
{
	unsigned int		irq_num;		/* IRQ like */
	unsigned long		base_addr;		/* base address where registers are */
	/* below two registers are for local bus */
	unsigned int		base_offs;		/* base address offset if necessary */
	unsigned int		data_offs;		/* data address offset if necessary */
} IRQMEM;

typedef struct _HWUSER
{
	HWPORT				*hwport;		/* reference back to the hardware port */
	unsigned int		index;			/* index to the user pool */
    unsigned int		xright;			/* read/write access right */
    unsigned int		running;		/* user in running or not */
	unsigned int		tx_timeout;		/* timeout for TX */
	unsigned int		rx_timeout;		/* timeout for RX */
} HWUSER;

/* contain the hardware information of a port */
struct _HWPORT
{
	unsigned int		port_idx;		/* index for this port */	
	unsigned char		*memp_addr;		/* memory map address */
	unsigned char		*base_addr;		/* I/O memory base address */
	unsigned char		*data_addr;		/* I/O memory data address */
	unsigned int		irq_num;		/* irq number */
	unsigned int		ier_reg;		/* interrupt enable register */
	unsigned int		ir_reg;			/* interrupt register */
	unsigned int		sr_reg;			/* status register */
	BUFFER				*rx_buf;		/* receiving buffer */
	BUFFER				*tx_buf;		/* transmiting buffer */
	unsigned int		running;		/* port in running or not */
	unsigned int		num_users;		/* number of users in the pool */
	HWUSER				*users[MAX_USERS_PER_PORT]; /* the pool of users for this port */
	wait_queue_head_t	wait_tx_event, wait_rx_event;	/* events for RX and TX access */
	spinlock_t			lock_base_addr, lock_open_count;/* locks for register access and user control */
	spinlock_t			lock_rx_buffer, lock_tx_buffer;	/* locks for RX and TX buffers */
	void				*private_data;
#ifdef WIN32
	wait_queue_head_t	wait_ir_event; /* for IST */
#endif
#ifdef PCI_BUS
	IRQMEM	irqm;
	char name[6];
#endif 
};

/* for each device, the following functions must be implemented */
unsigned int hwport_get_interrupts(HWPORT *port);
void hwport_purge_rxbuf(HWPORT *port);
void hwport_purge_txbuf(HWPORT *port);
void hwport_release(HWPORT*);
void hwport_handle_errors(HWPORT *port);
int hwport_read (HWPORT *port, char *buf, int size);
int hwport_write(HWPORT *port, char *data, int len);
int hwport_init(HWPORT*);
int	hwport_ioctl(HWPORT *port, unsigned int code, 
				 void *ibuf, unsigned int ilen, 
				 void *obuf, unsigned int olen, 
				 unsigned int *rlen);
void hwport_clear_interrupts(HWPORT *port, unsigned int type);

/* 
 * Wade: It is only read, not change any global variable.
 * So we don't need spin_lock_irqsave here 
 */
#define driver_port_get_rx_bytes(port, n) { \
	n = (port->rx_buf->datasize - port->rx_buf->consumed); \
}
#define driver_port_get_tx_bytes(port, n)  { \
	n = (port->tx_buf->datasize - port->tx_buf->consumed); \
}
#define driver_port_get_tx_space(port, n)  { \
	n = (port->tx_buf->datamax - port->tx_buf->datasize); \
}


/* the following functions are utility functions, they shall be valid for most drivers */

/* global function */

unsigned int driver_port_interrupt(HWPORT *port);
void driver_priavte_free(HWPORT *port);
int	driver_private_get_maxport(void);
int	driver_private_init(HWPORT *port, unsigned int idx);
int driver_private_irqmem(HWPORT *port, IRQMEM *irqm);
int driver_port_copy_to_user(HWPORT *port, char *buf, int max);
int driver_port_copy_from_user(HWPORT *port, char *data, int len);
int	driver_port_remove_user(HWUSER *user);
HWUSER* driver_port_create_user(HWPORT *port);
HWPORT*	driver_port_lookup(unsigned int port_idx);
#ifdef LOCAL_BUS
HWPORT*	driver_port_init(unsigned int port_idx, unsigned int mem_size);
#elif defined PCI_BUS
HWPORT*	driver_port_init(unsigned int port_idx, unsigned int mem_size, IRQMEM irqmem);
#endif
void driver_port_free(HWPORT *port);
void driver_port_free_all(void);
void driver_port_wake_up_user(HWUSER *user, unsigned int kill);
int driver_port_common_ioctl(HWUSER *user, unsigned int cmd, void *ibuf, unsigned int ilen, 
					void *obuf, unsigned int olen, unsigned int *rlen);

#ifdef __cplusplus
}
#endif

#endif
