/*  Copyright (C) MOXA Inc. All rights reserved.

    This software is distributed under the terms of the
    MOXA License.  See the file COPYING-MOXA for details.
*/
/*
    driver_port.c

    This defines some utility functions for generic drivers. 
	These functions should fit with many drivers.

    2009-10-15	CF Lin
		new release
*/
#include "driver.h"

static LIST *pList = NULL; /* the list of ports */

#ifdef WIN32
/*	to emulator a physical to virtual memory mapping in WIN32 
	Inputs:
		<pmem> physical
	Returns:
		the mapped area
*/
static unsigned char*
ioremap(unsigned long pmem, unsigned int size)
{
	unsigned char *mem_map;

	(void) size; /* igonre */

	/* generate a memory area */
	mem_map = (LPVOID) VirtualAlloc(0, PAGE_SIZE, MEM_RESERVE, PAGE_NOACCESS);
	if (mem_map==NULL)
		return NULL;
	if (VirtualCopy((LPVOID)mem_map, (LPVOID)(pmem), 
		PAGE_SIZE, PAGE_READWRITE | PAGE_NOCACHE) == FALSE)
	{
		VirtualFree(mem_map, PAGE_SIZE, MEM_RELEASE);
		mem_map = NULL;
	}
	return mem_map;
}

/*	to emulator a  virtual memory unmapping in WIN32 
	Inputs:
		<mem_map> mapped memory
	Returns:
		None
*/
static void
iounmap(unsigned char *mem_map)
{
	if (mem_map) VirtualFree(mem_map, PAGE_SIZE, MEM_RELEASE);
}

/* post action for free a port 
	Inputs:
		<port> the port
	Returns:
		None
*/
static void
driver_port_post_free(HWPORT *port)
{
	/* free up handles */
	if (port->wait_tx_event)	CloseHandle(port->wait_tx_event);
	if (port->wait_rx_event)	CloseHandle(port->wait_rx_event);

	/* free CS */
	DeleteCriticalSection(&port->lock_rx_buffer);
	DeleteCriticalSection(&port->lock_tx_buffer);
	DeleteCriticalSection(&port->lock_open_count);
	DeleteCriticalSection(&port->lock_base_addr);
	/* wake up the IST thread */
	if (port->wait_ir_event)
	{
		HANDLE temp = port->wait_ir_event;

		SetEvent(port->wait_ir_event);
		port->wait_ir_event = NULL;
		/* intentionally wait for IST to get out the infinite loop */
		msleep(2000);
		CloseHandle(port->wait_ir_event);
	}
}

#else /* WIN32 */

/*	post action for free a port 
	Inputs:
		<port> the port
	Returns:
		None
*/
static void
driver_port_post_free(HWPORT *port)
{
	if (port->irq_num)	free_irq(port->irq_num, port);
}

#endif /* WIN32 */

/*	handle interrupts for a hardware port 
	Inputs:
		<port> the port
		<interrupt> poistive value for intentional actions
	Returns:
		the combination of interrupt types
		
*/
unsigned int
driver_port_interrupt(HWPORT *port)
{
	unsigned int interrupt;

	interrupt = hwport_get_interrupts(port);

	if (interrupt==0)
		return 0;

	/* 
	 * XXX: remove spin_lock in ISR, 
	 * this is because no the same port ISR will be can at the same time twice
	 * so we only need to use spinlock_irqsave to protect tx / rx buf
	 * in driver_port_copy_to_user / driver_port_copy_from_user
	 */
	if (interrupt & INTERRUPT_TYPE_RX) /* handle RX */
	{
		/* protecting the rx_buf stuff */
 		//spin_lock(&port->lock_rx_buffer);
		/* call hardware action to queue data into the rx_buf */
		buffer_queue(port->rx_buf, (void*) port, 0, (buffer_access_t) hwport_read);
		/* leave protection */
 		//spin_unlock(&port->lock_rx_buffer);
		/* wake up users who waited for rx data */
		wake_up_interruptible(&port->wait_rx_event);
	}
	if (interrupt & INTERRUPT_TYPE_TX) /* handle TX */
	{
		/* protecting the tx_buf stuff */
 		//spin_lock(&port->lock_tx_buffer);
		/* write buffered data to the device */
		buffer_popup(port->tx_buf, (void*) port, 0, (buffer_access_t) hwport_write);
		/* leave protection */
 		//spin_unlock(&port->lock_tx_buffer);
		/* wake up users who waited for tx available */
		wake_up_interruptible(&port->wait_tx_event);
	}
	if (interrupt & INTERRUPT_TYPE_ERR) /* handle errors */
	{
		hwport_handle_errors(port);
		wake_up_interruptible(&port->wait_rx_event);
		wake_up_interruptible(&port->wait_tx_event);
	}
	return interrupt;
}

/*	copy data from the user buffer to the tx_buf and 
	then write buffered to the hardware port
	Inputs:
		<port> the port
		<data> user data
		<len> the lenght of the data
	Returns:
		the # of bytes being copied
*/
int
driver_port_copy_from_user(HWPORT *port, char *buf, int len)
{
	unsigned long flags=0;

	/* protecting the tx_buf stuff */
	spin_lock_irqsave(&port->lock_tx_buffer, flags);

	/* copy data to available space */
	if (len) len = buffer_queue(port->tx_buf, (void*) buf, len, (buffer_access_t) copy_from_user);
	/* write buffered data to the device */
	buffer_popup (port->tx_buf, (void*) port, 0, (buffer_access_t) hwport_write);

	/* leave protection */
	spin_unlock_irqrestore(&port->lock_tx_buffer, flags);

	return len;
}

#ifndef WIN32
static	char BUF[2048];
#endif

/*	copy data from the rx_buf to the user buffer
	Inputs:
		<port> the port
		<buf> the user buffer
		<max> the lenght of the buffer
	Returns:
		the # of bytes being copied
*/
int
driver_port_copy_to_user(HWPORT *port, char *buf, int max)
{
	unsigned long flags=0;
	int len;

	/* protecting the rx_buf stuff */
	spin_lock_irqsave(&port->lock_rx_buffer, flags);
	/* copy data from the rx_buf to the user buffer */
	len = buffer_popup (port->rx_buf, (void*) buf, max, (buffer_access_t) copy_to_user);
	if (len==0)
	{
		/* read data from the port directly */
#ifndef WIN32
		len = hwport_read(port, BUF, max);
		if (len) copy_to_user(buf, BUF, len);
#else
		len = hwport_read(port, buf, max);
#endif
	}
	/* leave protection */
	spin_unlock_irqrestore(&port->lock_rx_buffer, flags);
	return len;
}

/*	create a user over a hardware port 
	Inputs:
		<port> the port
	Returns:
		a user, otherwise NULL
*/
HWUSER*
driver_port_create_user(HWPORT *port)
{
	HWUSER *user = NULL;

	/* protect multiple access to the user count */
	spin_lock(&port->lock_open_count);
	if (port->num_users < MAX_USERS_PER_PORT)
	{
		user = (HWUSER*) MALLOC(sizeof(HWUSER));
		if (user)
		{
			memset(user, 0 , sizeof(HWUSER));
			/* each user carries its port info */
			user->hwport = port;
			/* add a user into the user pool */
			port->users[port->num_users] = user;
			user->index = port->num_users++;
			user->tx_timeout = DEFAULT_TIMEOUT;
			user->rx_timeout = DEFAULT_TIMEOUT;
			user->running = 1;
			if (port->num_users==1) /* trigger the chip */
			{
				hwport_init(port);
			}
		}
	}
	/* leave the protection */
	spin_unlock(&port->lock_open_count);
	return user;
}

/* wake up all users who are waiting for rx and tx data
	Inputs:
		<user> the user
		<kill> indicating if the caller intentionally kills all
	Returns:
		0 on success
*/
void
driver_port_wake_up_user(HWUSER *user, unsigned int kill)
{
	HWPORT *port;

	if (!user || (port = user->hwport)==NULL)
		return;
	user->running = 0;
	/*	read/write threads of this user may be blocked at this moment,
		wakeup them	*/
	wake_up_interruptible(&port->wait_rx_event);
	wake_up_interruptible(&port->wait_tx_event);
	if (kill)
	{
		msleep(2000);
	}
}

/*	remove a user from a hardware port 
	Inputs:
		<user> the user
	Returns:
		0 on success
*/
int
driver_port_remove_user(HWUSER *user)
{
	HWPORT *port;

	if (!user || (port = user->hwport)==NULL || port->num_users==0)
		return -1;
	/* make sure we don't collide with an in-progress open */
	spin_lock(&port->lock_open_count);
	/* decrease # of users */
	port->num_users--;
	if (port->num_users==0) /* ohoh, no user */
	{
		hwport_release(port);
	}
	/* remove the user from the pool */
	if (user->index != port->num_users)
	{
		/* if it is not the last one, change the position */
		port->users[user->index] = port->users[port->num_users];
	}
	/* leave the protection */
	spin_unlock(&port->lock_open_count);
	driver_port_wake_up_user(user, 0);
	memset(user, 0, sizeof(HWUSER));
	FREE(user);
	return 0;
}

/*	lookup a hardware port by an index
	Inputs:
		<idx> the index to the pool of ports maintained by pList
	Returns:
		the associated port on success, otherwise return NULL
*/
HWPORT*
driver_port_lookup(unsigned int idx)
{
	LIST *list;
	HWPORT *port;

	/* walk through the list */
	for (list=pList; list; list=list->next)
	{
		port = (HWPORT*) list->data;
		if (port->port_idx==idx) return port;
	}
	return NULL;
}

/*	a tricky one to indicate if the caller want to remove all ports
	Note : to avoid core dump
*/
static int global_remove_all = 0;

/*	free up a hardware port
	Inputs:
		<port> the port
	Returns:
		None
*/
void
driver_port_free(HWPORT *port)
{
	dbg_function_in();

	if (!port)
		return;

	port->running = 0;

	/* release the hardware part */
	hwport_release(port);
	/* release users one by one, from the last one */
	while(port->num_users)
	{
		driver_port_remove_user(port->users[port->num_users-1]);
	}
	/* free up allocated buffers */
	if (port->tx_buf)	buffer_free(port->tx_buf);
	if (port->rx_buf)	buffer_free(port->rx_buf);
	/* remove it from the list */
	if (global_remove_all==0)
	{
		/* free the list container */
		pList = list_delete(pList, port);
	}
	/* free up mapped memory */
#ifdef LOCAL_BUS
	if (port->memp_addr) iounmap(port->memp_addr);
#elif defined PCI_BUS
	/* release io port address (not memp addr)... */
	if (port->memp_addr) ioport_unmap(port->memp_addr);
	/* release_region */
	release_region(port->irqm.base_addr, 128);
#endif
	/* for diffferent OSs */
	driver_port_post_free(port); /* free irq here */
	driver_priavte_free(port);
	/* NULL everything */
	memset(port, 0, sizeof(HWPORT));
	FREE(port);
}

/*	free up all hardware ports
	Inputs:
		None
	Returns:
		None
*/
void
driver_port_free_all(void)
{
	dbg_function_in();
	global_remove_all = 1; /* tricky one */
	list_free(pList, (list_destroy_t) driver_port_free);
	global_remove_all = 0; /* tricky one */
	pList = NULL;
}

/*	initialize a hardware port
	Inputs:
		<port_idx> the index of the port
		<mem_size> memory size to be callocated for the register addresses
	Returns:
		the port if successful, otherwise return NULL
*/
#ifdef LOCAL_BUS
HWPORT*
driver_port_init(unsigned int port_idx, unsigned int mem_size)
{
	HWPORT *port;
	IRQMEM irqm;

	dbg_function_in(); /* debug */
	/* allocate memory for each port */
	port = MALLOC(sizeof(HWPORT));
	if (!port)
		return NULL;
	memset(port, 0, sizeof(HWPORT));
	port->running = 1;
	port->port_idx = port_idx;
	port->rx_buf = buffer_new(MSG_UNIT_SIZE*MAX_BUF_MSG);
	port->tx_buf = buffer_new(MSG_UNIT_SIZE*MAX_BUF_MSG);
	if (!port->rx_buf || !port->tx_buf)
		goto err;
	/* possible further actions, and initial port's private data */
	if (driver_private_init(port, port_idx))
		goto err;
	/* get IRQ and base memory */
	if (driver_private_irqmem(port, &irqm))
		goto err;
	/* physical I/O memory to virtual memory */
	port->memp_addr = ioremap(irqm.base_addr, mem_size);
	if (port->memp_addr==NULL)
		goto err;
	/* locate register addresses before initialize the hardware */
	port->base_addr = port->memp_addr + irqm.base_offs;
	port->data_addr = port->memp_addr + irqm.data_offs;
	port->irq_num = irqm.irq_num;
	/* create critical sections : open, rx_buf, tx_buf */
	spin_lock_init(&port->lock_open_count);
	spin_lock_init(&port->lock_rx_buffer);
	spin_lock_init(&port->lock_tx_buffer);
	/* Initialize the critical section that will guard hardward read/write. */
	spin_lock_init(&port->lock_base_addr);
	/* initialize tx/rx wait queue/event */
	init_waitqueue_head(&port->wait_tx_event);
	init_waitqueue_head(&port->wait_rx_event);
#ifdef WIN32
	if (port->wait_rx_event==NULL || port->wait_tx_event==NULL) 
		goto err;
#endif
	/* hook the port in the global list */
	pList = list_push_data(pList, port);
	if (pList)
		return port;
err:
	if (port) driver_port_free(port);
	return NULL;
}

#elif defined PCI_BUS

HWPORT*
driver_port_init(unsigned int port_idx, unsigned int mem_size, IRQMEM irqm)
{
	HWPORT *port;

	dbg_function_in(); /* debug */
	/* allocate memory for each port */
	port = MALLOC(sizeof(HWPORT));
	if (!port)
		return NULL;
	memset(port, 0, sizeof(HWPORT));
	port->running = 1;
	port->port_idx = port_idx;
	port->rx_buf = buffer_new(MSG_UNIT_SIZE*MAX_BUF_MSG);
	port->tx_buf = buffer_new(MSG_UNIT_SIZE*MAX_BUF_MSG);
	if (!port->rx_buf || !port->tx_buf)
		goto err;

	/* possible further actions, and initial port's private data */
	if (driver_private_init(port, port_idx)) /* malloc port->private_data */
		goto err;

	/* get IRQ and base memory */
	port->irqm = irqm;

	/* 
	 * request region for io port exclusive 
	 * note: for io port (not the mem_addr after ioport_map
	 */
	if (!request_region(port->irqm.base_addr, 128, "can")) {
		printk("error: request region fail, base_addr=%x\n", (unsigned int)irqm.base_addr);
		goto err;
	}

	/* create pseudo address from io port */
	port->memp_addr = ioport_map(port->irqm.base_addr, 128);
	if (port->memp_addr==NULL)
		goto err;

	/* locate register addresses before initialize the hardware */
	port->base_addr = port->memp_addr;
	port->irq_num = port->irqm.irq_num;

	/* FIXME: set IER = 0 until port be open*/
	hwport_clear_interrupts(port, INTERRUPT_TYPE_RX|INTERRUPT_TYPE_TX|INTERRUPT_TYPE_ERR);

	/* create critical sections : open, rx_buf, tx_buf */
	spin_lock_init(&port->lock_open_count);
	spin_lock_init(&port->lock_rx_buffer);
	spin_lock_init(&port->lock_tx_buffer);
	/* Initialize the critical section that will guard hardward read/write. */
	spin_lock_init(&port->lock_base_addr);
	/* initialize tx/rx wait queue/event */
	init_waitqueue_head(&port->wait_tx_event);
	init_waitqueue_head(&port->wait_rx_event);
	/* hook the port in the global list */
	pList = list_push_data(pList, port);
	if (pList)
		return port;
err:
	if (port) driver_port_free(port);
	return NULL;
}
#endif 
/*	define some common ioctl here 
	Inputs:
		<user> the user to access ioctl
		<code> command code
		<ibuf> the input buffer
		<ilen> the length of the input buffer in WIN32, 0 in Linux
		<obuf> the output buffer (in Linux, it is same as ibuf)
		<olen> the length of the output buffer in WIN32, 0 in Linux
		<rlen> the # of bytes to be returned (NULL in Linux)
	Returns:
		0 on success
*/
int
driver_port_common_ioctl(HWUSER *user, unsigned int code, 
					void *ibuf, unsigned int ilen, 
					void *obuf, unsigned int olen,
					unsigned int *rlen)
{
	HWPORT	*port;
	unsigned int ival, size = 0;
	int ret = 0;

	port = user->hwport;
	switch ( code )
	{
	/* put it on the 1st item */
	case IOCTL_GET_RX_BYTES: /* return the # of bytes in the rx buffer */
		size = sizeof(unsigned int);
		if (!access_ok(VERIFY_WRITE, obuf, size))
			ret = -2;
		else
		{
			driver_port_get_rx_bytes(port, ival);
			copy_to_user((unsigned int*) obuf, (void *) &ival, size);
		}
		break;
	case IOCTL_GET_TX_BYTES: /* return the # of bytes in the tx buffer */
		size = sizeof(unsigned int);
		if (!access_ok(VERIFY_WRITE, obuf, size))
			ret = -2;
		else
		{
			driver_port_get_tx_bytes(port, ival);
			copy_to_user((unsigned int*) obuf, (void *) &ival, size);
		}
		break;
	case IOCTL_PURGE_RX_BUFFER:
		/*  purge rx_buf and the data in the hardware port */
		buffer_reset(port->rx_buf);
		hwport_purge_rxbuf(port);
		break;
	case IOCTL_PURGE_TX_BUFFER:
		/*  purge tx_buf and the data in the hardware port */
		buffer_reset(port->tx_buf);
		hwport_purge_txbuf(port);
		break;
	case IOCTL_SET_READ_TIMEOUT: /* set time out in ms for reading */
		size = sizeof(unsigned int);
		if( !access_ok(VERIFY_READ, ibuf, size)) 
			ret = -3;
		else
			copy_from_user( (void *) &user->rx_timeout, (unsigned int*) ibuf, size);
		break;
	case IOCTL_SET_WRITE_TIMEOUT: /* set time out in ms for writing */
		size = sizeof(unsigned int);
		if( !access_ok(VERIFY_READ, ibuf, size)) 
			ret = -3;
		else
			copy_from_user( (void *) &user->tx_timeout, (unsigned int*) ibuf, size);
		break;
	default:
		ret = -99;
		break;
	}
	if (rlen) *rlen = size;
	return ret;
}

